!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2010.  Los Alamos National Security, LLC. This material was    !
! produced under U.S. Government contract DE-AC52-06NA25396 for Los Alamos !
! National Laboratory (LANL), which is operated by Los Alamos National     !
! Security, LLC for the U.S. Department of Energy. The U.S. Government has !
! rights to use, reproduce, and distribute this software.  NEITHER THE     !
! GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY,     !
! EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS         !
! SOFTWARE.  If software is modified to produce derivative works, such     !
! modified software should be clearly marked, so as not to confuse it      !
! with the version available from LANL.                                    !
!                                                                          !
! Additionally, this program is free software; you can redistribute it     !
! and/or modify it under the terms of the GNU General Public License as    !
! published by the Free Software Foundation; version 2.0 of the License.   !
! Accordingly, this program is distributed in the hope that it will be     !
! useful, but WITHOUT ANY WARRANTY; without even the implied warranty of   !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General !
! Public License for more details.                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE BOEVECS
  
  USE CONSTANTS_MOD
  USE SETUPARRAY
  USE DIAGARRAY
  USE MYPRECISION

  IMPLICIT NONE

  INTEGER :: I, J, K
  INTEGER :: ITER, BREAKLOOP
  REAL(LATTEPREC) :: OCCTARGET, OCC, FDIRAC, DFDIRAC
  REAL(LATTEPREC) :: OCCERROR, SHIFTCP, FDIRACARG
  REAL(LATTEPREC) :: PREVERROR, PREVERROR2, PREVERROR3
  REAL(LATTEPREC) :: MAXSHIFT = ONE

  BO = ZERO

  OCCTARGET = BNDFIL*FLOAT(HDIM)
  
  ITER = 0

  BREAKLOOP = 0

  OCCERROR = ZERO
  PREVERROR = ZERO
  PREVERROR2 = ZERO
  PREVERROR3 = ZERO

  !
  ! The do-while loop uses a Newton-Raphson optimization of the chemical 
  ! potential to obtain the correct occupation
  !

  IF (KBT .GT. 0.000001) THEN  ! This bit is for a finite electronic temperature

     DO WHILE  ( BREAKLOOP .EQ. 0 )

        ITER = ITER + 1
        
        IF (ITER .EQ. 100) THEN
           WRITE(6,*) "NR scheme to find the Chemical potential is broken"
           STOP
        ENDIF
        
        OCC = ZERO
        DFDIRAC = ZERO
        
        DO I = 1, HDIM
           
           FDIRACARG = (EVALS(I) - CHEMPOT)/KBT
           
           IF (FDIRACARG .GT. 46.0D0) THEN
              FDIRAC = ZERO
           ELSEIF (FDIRACARG .LT. -46.0D0) THEN
              FDIRAC = ONE
           ELSE
              FDIRAC = ONE/(ONE + EXP(FDIRACARG))
           ENDIF
           
           OCC = OCC + FDIRAC
           
           DFDIRAC = DFDIRAC + EXP(FDIRACARG)*FDIRAC*FDIRAC/KBT
           
        ENDDO
        
        PREVERROR3 = PREVERROR2
        PREVERROR2 = PREVERROR
        PREVERROR = OCCERROR
        
        OCCERROR = OCCTARGET - OCC
        
        SHIFTCP = OCCERROR/DFDIRAC
        
        IF (ABS(SHIFTCP) .GT. MAXSHIFT) THEN
           SHIFTCP = SIGN(MAXSHIFT, SHIFTCP)
        ENDIF
        
        CHEMPOT = CHEMPOT + SHIFTCP
        
!             PRINT*, ITER, OCCERROR, PREVERROR, PREVERROR2, PREVERROR3
        
        IF (LATTEPREC .EQ. KIND(0.0D0)) THEN
           
        ! Convergence criterion for double precision...
           
           IF (ABS(OCCERROR) .LT. BREAKTOL) THEN
              BREAKLOOP = 1
           ENDIF
           
        ELSE
        
           ! ...and for single precison
           
           IF (OCCERROR .EQ. PREVERROR .OR. &
                OCCERROR .EQ. PREVERROR2 .OR. &
                OCCERROR .EQ. PREVERROR3 .OR. ITER .EQ. 25) THEN
              
              BREAKLOOP = 1
              
           ENDIF
           
        ENDIF
        
     ENDDO
     
     ! Now we have the chemical potential we can build the density matrix

     DO I = 1, HDIM
        
        FDIRACARG = (EVALS(I) - CHEMPOT)/KBT
        
        IF (FDIRACARG .GT. 46.0D0) THEN
           FDIRAC = ZERO
        ELSEIF (FDIRACARG .LT. -46.0D0) THEN
           FDIRAC = ONE
        ELSE
           FDIRAC = ONE/(ONE + EXP(FDIRACARG))
        ENDIF
        
        DO J = 1, HDIM
           
           SCALEDEVECS(J,I) = EVECS(J,I)*FDIRAC
           
        ENDDO
        
     ENDDO

!$OMP PARALLEL DO DEFAULT(NONE) SCHEDULE(GUIDED) &
!$OMP SHARED(HDIM, BO, SCALEDEVECS, EVECS) &
!$OMP PRIVATE(I, J, K)
     
     DO I = 1, HDIM
        DO J = 1, HDIM
           DO K = 1, HDIM
              
              BO(J,I) = BO(J,I) + SCALEDEVECS(J,K)*EVECS(I,K)
              
           ENDDO
        ENDDO
     ENDDO

!$OMP END PARALLEL DO

  ELSE ! This bit is for zero electronic temperature

     !
     ! This definition of the chemical potential is a little arbitrary
     !

     CHEMPOT = HALF*(EVALS(INT(OCCTARGET)) + EVALS(INT(OCCTARGET) + 1))

!$OMP PARALLEL DO DEFAULT(NONE) SCHEDULE(GUIDED) &
!$OMP SHARED(HDIM, OCCTARGET, BO, EVECS) &
!$OMP PRIVATE(I, J, K)

     DO I = 1, HDIM
        DO J = 1, HDIM
           DO K = 1, INT(OCCTARGET)

              BO(J,I) = BO(J,I) + EVECS(J,K)*EVECS(I,K)

           ENDDO
        ENDDO
     ENDDO

!$OMP END PARALLEL DO
           
  ENDIF

  BO = TWO * BO
  
  RETURN
  
END SUBROUTINE BOEVECS

