!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2010.  Los Alamos National Security, LLC. This material was    !
! produced under U.S. Government contract DE-AC52-06NA25396 for Los Alamos !
! National Laboratory (LANL), which is operated by Los Alamos National     !
! Security, LLC for the U.S. Department of Energy. The U.S. Government has !
! rights to use, reproduce, and distribute this software.  NEITHER THE     !
! GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY,     !
! EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS         !
! SOFTWARE.  If software is modified to produce derivative works, such     !
! modified software should be clearly marked, so as not to confuse it      !
! with the version available from LANL.                                    !
!                                                                          !
! Additionally, this program is free software; you can redistribute it     !
! and/or modify it under the terms of the GNU General Public License as    !
! published by the Free Software Foundation; version 2.0 of the License.   !
! Accordingly, this program is distributed in the hope that it will be     !
! useful, but WITHOUT ANY WARRANTY; without even the implied warranty of   !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General !
! Public License for more details.                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE FERMIEXPANS

  USE CONSTANTS_MOD
  USE SETUPARRAY
  USE FERMICOMMON
  USE SPINARRAY
  USE MYPRECISION

  IMPLICIT NONE

  INTEGER :: I, J, ITER
  INTEGER :: BREAKLOOP
  REAL(LATTEPREC) :: OCC, OCCERROR
  REAL(LATTEPREC) :: PREVERROR, PREVERROR2, PREVERROR3
  REAL(LATTEPREC) :: BOVER2M, TRX, TRXOMX
  REAL(LATTEPREC) :: SHIFTCP, MAXSHIFT = 1.0D0

  OCC = BNDFIL*FLOAT(HDIM)

  ITER = 0

  BREAKLOOP = 0

  BOVER2M = ONE/(KBT*(TWO**(2 + FERMIM)))

  SCFS = 0

  PREVERROR = ZERO
  PREVERROR2 = ZERO
  PREVERROR3 = ZERO

  DO WHILE (BREAKLOOP .EQ. 0)

     SCFS = SCFS + 1

     ITER = ITER + 1

     IF (ITER .EQ. 100) THEN
        WRITE(6,*) "Fermi expansion is not converging: STOP!"
        CALL PANIC
        STOP
     ENDIF     

     IF (SPINON .EQ. 0) THEN

        DO I = 1, HDIM
           DO J = I, HDIM
              
              IF ( I .EQ. J ) THEN
                 
                 BO(I,I) = HALF - BOVER2M*(H(I,I) - CHEMPOT)
                 
              ELSE
                 
                 BO(J,I) = MINUSONE*BOVER2M*H(J,I)
                 BO(I,J) = BO(J,I)
                 
              ENDIF
              
           ENDDO
        ENDDO

     ELSE 

        DO I = 1, HDIM
           DO J = I, HDIM
              
              IF ( I .EQ. J ) THEN
                 
                 RHOUP(I,I) = HALF - BOVER2M*(HUP(I,I) - CHEMPOT)
                 RHODOWN(I,I) = HALF - BOVER2M*(HDOWN(I,I) - CHEMPOT)
                 
              ELSE
                 
                 RHOUP(J,I) = MINUSONE*BOVER2M*HUP(J,I)
                 RHOUP(I,J) = RHOUP(J,I)

                 RHODOWN(J,I) = MINUSONE*BOVER2M*HDOWN(J,I)
                 RHODOWN(I,J) = RHODOWN(J,I)

              ENDIF
              
           ENDDO
        ENDDO

     ENDIF

     DO I = 1, FERMIM

        IF (CGORLIB .EQ. 0) THEN
           CALL SOLVEMATLAPACK
        ELSE
           CALL SOLVEMATCG
        ENDIF

     ENDDO

     ! Modifying chemical potential

     TRX = ZERO
     TRXOMX = ZERO

     IF (SPINON .EQ. 0) THEN

        DO I = 1, HDIM
           DO J = I, HDIM
              
              IF (I .EQ. J) THEN
                 
                 TRXOMX = TRXOMX + BO(I,I)*(ONE - BO(I,I))
                 
              ELSE
                 
                 TRXOMX = TRXOMX - TWO*BO(J,I)*BO(J,I)
                 
              ENDIF
              
           ENDDO
           
           TRX = TRX + BO(I,I)

        ENDDO

        SHIFTCP = KBT*(OCC - TRX)/TRXOMX
        
        PREVERROR3 = PREVERROR2
        PREVERROR2 = PREVERROR
        PREVERROR = OCCERROR
        OCCERROR = ABS(OCC - TRX)
!        PRINT*, ITER, OCCERROR

     ELSE 

        DO I = 1, HDIM
           DO J = I, HDIM
              
              IF (I .EQ. J) THEN
                 
                 TRXOMX = TRXOMX + RHOUP(I,I)*(ONE - RHOUP(I,I)) + &
                      RHODOWN(I,I)*(ONE - RHODOWN(I,I))
                 
              ELSE
                 
                 TRXOMX = TRXOMX - TWO*(RHOUP(J,I)*RHOUP(J,I) + &
                      RHODOWN(J,I)*RHODOWN(J,I))
                 
              ENDIF
              
           ENDDO

           TRX = TRX + RHOUP(I,I) + RHODOWN(I,I)

        ENDDO

        SHIFTCP = KBT*(TOTNE - TRX)/TRXOMX
        
        PREVERROR3 = PREVERROR2
        PREVERROR2 = PREVERROR
        PREVERROR = OCCERROR
        OCCERROR = ABS(TOTNE - TRX)        

     ENDIF

!     PRINT*, ITER, OCCERROR

     IF (ABS(SHIFTCP) .GT. MAXSHIFT) THEN
        SHIFTCP = SIGN(MAXSHIFT, SHIFTCP)
     ENDIF

     CHEMPOT = CHEMPOT + SHIFTCP

     IF (LATTEPREC .EQ. KIND(0.0D0)) THEN

        IF (ITER .GE. 3 .AND. OCCERROR .LT. BREAKTOL) THEN
           BREAKLOOP = 1
        ENDIF

     ELSE 

        IF ( ITER .GE. 3 .AND. ((OCCERROR .LT. BREAKTOL) .AND. &
             (OCCERROR .EQ. PREVERROR .OR. &
             OCCERROR .EQ. PREVERROR2 .OR. &
             OCCERROR .EQ. PREVERROR3)) .OR. & 
             ITER .EQ. 10) THEN
           
           BREAKLOOP = 1
        
        ENDIF

     ENDIF

  ENDDO

  IF (SPINON .EQ. 0) THEN
     BO = TWO*BO
  ENDIF

  RETURN

END SUBROUTINE FERMIEXPANS
           
  
  
