!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2010.  Los Alamos National Security, LLC. This material was    !
! produced under U.S. Government contract DE-AC52-06NA25396 for Los Alamos !
! National Laboratory (LANL), which is operated by Los Alamos National     !
! Security, LLC for the U.S. Department of Energy. The U.S. Government has !
! rights to use, reproduce, and distribute this software.  NEITHER THE     !
! GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY,     !
! EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS         !
! SOFTWARE.  If software is modified to produce derivative works, such     !
! modified software should be clearly marked, so as not to confuse it      !
! with the version available from LANL.                                    !
!                                                                          !
! Additionally, this program is free software; you can redistribute it     !
! and/or modify it under the terms of the GNU General Public License as    !
! published by the Free Software Foundation; version 2.0 of the License.   !
! Accordingly, this program is distributed in the hope that it will be     !
! useful, but WITHOUT ANY WARRANTY; without even the implied warranty of   !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General !
! Public License for more details.                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE INITSPARSESP2

  USE CONSTANTS_MOD
  USE SETUPARRAY
  USE PUREARRAY
  USE SPARSEIND
  USE MYPRECISION

  IMPLICIT NONE

  INTEGER :: I, J, L, M, ITER
  INTEGER :: SUMTOT, PREVSUMTOT, DIFFTOT
  INTEGER, ALLOCATABLE :: TMP(:,:)
  REAL(LATTEPREC) ::  TRX, OCC
  REAL(LATTEPREC) ::  TRX2, TR2XX2, LIMIT1, LIMIT2
  REAL(LATTEPREC) :: TOL

  TOL = 1.0D-6

  OCC = BNDFIL*DFLOAT(HDIM)

  TRX = ZERO

  DO I = 1, HDIM
     DO J = I, HDIM

        IF (I .EQ. J) THEN

           BO(I,I) = (MAXEVAL - H(I,I))/MAXMINUSMIN

           TRX = TRX + BO(J,I)

        ELSE

           BO(J,I) = (ZERO - H(J,I))/MAXMINUSMIN
           BO(I,J) = BO(J,I)

        ENDIF

     ENDDO
  ENDDO

  ITER = 0

  SUMTOT = 0
  DIFFTOT = 1

  DO WHILE (DIFFTOT .NE. 0)

     PREVSUMTOT = SUMTOT
     SUMTOT = 0

     ITER = ITER + 1

     IF (LATTEPREC .EQ. KIND(0.0D0)) THEN
        CALL DGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0D0, &
             BO, HDIM, BO, HDIM, 0.0D0, X2, HDIM)
     ELSE 
        CALL SGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0, &
             BO, HDIM, BO, HDIM, 0.0, X2, HDIM)
     ENDIF

     TRX2 = ZERO
     TR2XX2 = ZERO

!     TWOXX2 = TWO*BO - X2

     DO I = 1, HDIM
        TRX2 = TRX2 + X2(I,I)
        TR2XX2 = TR2XX2 + TWO*BO(I,I) - X2(I,I)
     ENDDO

     LIMIT1 = ABS(TRX2 - OCC)
     LIMIT2 = ABS(TR2XX2 - OCC)
     
     IF (LIMIT1 .LT. LIMIT2) THEN

        BO = X2

     ELSEIF  (LIMIT2 .LT. LIMIT1) THEN

        BO = TWO*BO - X2

     ENDIF

  
  !
  ! We are now able to figure out the sparsity of the matrix X
  !

     TOTNONZERO = 0

     DO I = 1, HDIM
        DO J = 1, HDIM

           IF (ABS(BO(I,J)) .GT. TOL) THEN
              
              TOTNONZERO(I) = TOTNONZERO(I) + 1

              IF (TOTNONZERO(I) .GT. MAXDIM) THEN
                 
!              PRINT*, "Increasing size of sparse map array"

                 ALLOCATE(TMP(HDIM, MAXDIM))
                 
                 MAXDIM = TOTNONZERO(I)
                 
                 DO L = 1, I
                    DO M = 1, MAXDIM - 1
                       TMP(L,M) = NONZERO(L,M)
                    ENDDO
                 ENDDO
                 
                 DEALLOCATE (NONZERO)
                 
                 ALLOCATE(NONZERO(HDIM,MAXDIM))
                 
                 DO L = 1, I
                    DO M = 1, MAXDIM - 1
                       NONZERO(L,M) = TMP(L,M)
                    ENDDO
                 ENDDO
                 
                 DEALLOCATE(TMP)

              ENDIF

              NONZERO(I,TOTNONZERO(I)) = J
           
           ENDIF

        ENDDO

        SUMTOT = SUMTOT + TOTNONZERO(I)

     ENDDO

     DIFFTOT = SUMTOT - PREVSUMTOT

  ENDDO

  RETURN

END SUBROUTINE INITSPARSESP2
