!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2010.  Los Alamos National Security, LLC. This material was    !
! produced under U.S. Government contract DE-AC52-06NA25396 for Los Alamos !
! National Laboratory (LANL), which is operated by Los Alamos National     !
! Security, LLC for the U.S. Department of Energy. The U.S. Government has !
! rights to use, reproduce, and distribute this software.  NEITHER THE     !
! GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY,     !
! EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS         !
! SOFTWARE.  If software is modified to produce derivative works, such     !
! modified software should be clearly marked, so as not to confuse it      !
! with the version available from LANL.                                    !
!                                                                          !
! Additionally, this program is free software; you can redistribute it     !
! and/or modify it under the terms of the GNU General Public License as    !
! published by the Free Software Foundation; version 2.0 of the License.   !
! Accordingly, this program is distributed in the hope that it will be     !
! useful, but WITHOUT ANY WARRANTY; without even the implied warranty of   !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General !
! Public License for more details.                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE SP2PURE

  !
  ! This subroutine implements Niklasson's SP2 density matrix purification
  ! algorithm. 
  !

  USE CONSTANTS_MOD
  USE SETUPARRAY
  USE PUREARRAY
  USE SPINARRAY
  USE MYPRECISION

  IMPLICIT NONE

  INTEGER :: I, J, ITER
  INTEGER :: BREAKLOOP
  REAL(LATTEPREC) :: TRX, OCC
  REAL(LATTEPREC) :: TRX2, TR2XX2, LIMIT1, LIMIT2
  
  !
  ! We're also using Niklasson's scheme to determine convergence
  !

  REAL(LATTEPREC) :: IDEMPERR0, IDEMPERR1, IDEMPERR2, TRXOLD  

  IDEMPERR0 = ZERO
  IDEMPERR1 = ZERO 
  IDEMPERR2 = ZERO

  OCC = BNDFIL*FLOAT(HDIM)

  IF (SPINON .EQ. 0) THEN

     TRX = ZERO
     
     DO I = 1, HDIM
        DO J = I, HDIM
           
           IF (I .EQ. J) THEN
              
              BO(J,I) = (MAXEVAL - H(I,I))/MAXMINUSMIN
              
              TRX = TRX + BO(I,I)
              
           ELSE
              
              BO(J,I) = (ZERO - H(J,I))/MAXMINUSMIN
              BO(I,J) = BO(J,I)
              
           ENDIF
           
        ENDDO
     ENDDO

     ITER = 0
     
     BREAKLOOP = 0
     
     DO WHILE ( BREAKLOOP .EQ. 0 )
        
        ITER = ITER + 1
        
        IF (ITER .EQ. 100) THEN
           WRITE(6,*) "SP2 purification is not converging: STOP!"
           CALL PANIC
           STOP
        ENDIF
        
        IF (LATTEPREC .EQ. KIND(0.0D0)) THEN
           CALL DGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0D0, &
                BO, HDIM, BO, HDIM, 0.0D0, X2, HDIM)
        ELSE
           CALL SGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0, &
                BO, HDIM, BO, HDIM, 0.0, X2, HDIM)
        ENDIF
        
        TRX2 = ZERO
        TR2XX2 = ZERO
        
        DO I = 1, HDIM
           TRX2 = TRX2 + X2(I,I)
           TR2XX2 = TR2XX2 + TWO*BO(I,I) - X2(I,I)
        ENDDO
        
        LIMIT1 = ABS(TRX2 - OCC)
        LIMIT2 = ABS(TR2XX2 - OCC)
        
        IF (LIMIT1 .LT. LIMIT2) THEN
           
           BO = X2
           
        ELSE

           BO = TWO*BO - X2
           
        ENDIF
       
        TRXOLD = TRX

        TRX = ZERO        
        DO I = 1, HDIM
           TRX = TRX + BO(I,I)
        ENDDO

        IDEMPERR2 = IDEMPERR1
        IDEMPERR1 = IDEMPERR0
        IDEMPERR0 = ABS(TRX - TRXOLD)

!        PRINT*, ITER, IDEMPERR0 - IDEMPERR2, ABS(TRX - OCC)
!        PRINT*, ITER, ABS(TRX - OCC)
        
        IF ( ITER .GE. MINSP2ITER ) THEN
!           IF ( ABS(TRX - OCC) .LT. BREAKTOL ) THEN
           IF (ABS(TRX - OCC) .LT. BREAKTOL .AND. &
                IDEMPERR0 .GE. IDEMPERR2) THEN
              BREAKLOOP = 1
           ENDIF
        ENDIF

     ENDDO
     
     BO = TWO*BO

  ELSE

     !
     ! Start by remapping the two H matrices such that 
     ! all eigenvalues are [0:1]. We have the Gersgorin bounds
     ! for both Hup and Hdown
     !

     DO I = 1, HDIM
        DO J = I, HDIM
           
           IF (I .EQ. J) THEN
              
              RHOUP(J,I) = (MAXEVAL - HUP(J,I))/MAXMINUSMIN

              RHODOWN(J,I) = (MAXEVAL - HDOWN(J,I)) / &
                   MAXMINUSMIN

           ELSE
              
              RHOUP(J,I) = (ZERO - HUP(J,I))/MAXMINUSMIN
              RHOUP(I,J) = RHOUP(J,I)

              RHODOWN(J,I) = (ZERO - HDOWN(J,I))/MAXMINUSMIN
              RHODOWN(I,J) = RHODOWN(J,I)

           ENDIF
           
        ENDDO
     ENDDO

     ITER = 0

     BREAKLOOP = 0

     DO WHILE (BREAKLOOP .EQ. 0)

        ITER = ITER + 1

        IF (ITER .EQ. 100) THEN
           WRITE(6,*) "SP2 purification not converging: STOP!"
           CALL PANIC
           STOP
        ENDIF

        !
        ! Now we're calculating X*X for the spin-up and spin-down
        ! density matrices
        !

        IF (LATTEPREC .EQ. KIND(0.0D0)) THEN

           CALL DGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0D0, &
                RHOUP, HDIM, RHOUP, HDIM, 0.0D0, X2UP, HDIM)
           CALL DGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0D0, &
                RHODOWN, HDIM, RHODOWN, HDIM, 0.0D0, X2DOWN, HDIM)

        ELSE

           CALL SGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0, &
                RHOUP, HDIM, RHOUP, HDIM, 0.0, X2UP, HDIM)
           CALL SGEMM('N', 'N', HDIM, HDIM, HDIM, 1.0, &
                RHODOWN, HDIM, RHODOWN, HDIM, 0.0, X2DOWN, HDIM)

        ENDIF
        
        !
        ! Now we're getting the traces of the matrices to
        ! see which way to project
        !
       
        TRX2 = ZERO
        TR2XX2 = ZERO

        DO I = 1, HDIM

           TRX2 = TRX2 + X2UP(I,I) + X2DOWN(I,I)
           TR2XX2 = TR2XX2 + TWO*(RHOUP(I,I) + RHODOWN(I,I)) - &
                (X2UP(I,I) + X2DOWN(I,I))

        ENDDO        

        LIMIT1 = ABS(TRX2 - TOTNE)
        LIMIT2 = ABS(TR2XX2 - TOTNE)

        IF (LIMIT1 .LT. LIMIT2) THEN
           
           RHOUP = X2UP
           RHODOWN = X2DOWN

        ELSE
           
           RHOUP = TWO*RHOUP - X2UP
           RHODOWN = TWO*RHODOWN - X2DOWN
           
        ENDIF

        TRXOLD = TRX

        TRX = ZERO
        DO I = 1, HDIM
           TRX = TRX + RHOUP(I,I) + RHODOWN(I,I)
        ENDDO

        IDEMPERR2 = IDEMPERR1
        IDEMPERR1 = IDEMPERR0
        IDEMPERR0 = ABS(TRX - TRXOLD)

!        PRINT*, ITER, IDEMPERR0 - IDEMPERR2, ABS(TRX - TOTNE)

        IF ( ITER .GE. MINSP2ITER ) THEN
!           IF (ABS(TRX - TOTNE) .LT. BREAKTOL) THEN
           IF (ABS(TRX - TOTNE) .LT. BREAKTOL .AND. &
                IDEMPERR0 .GE. IDEMPERR2) THEN
              BREAKLOOP = 1
           ENDIF
        ENDIF

     ENDDO

  ENDIF

  RETURN

END SUBROUTINE SP2PURE
