!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2010.  Los Alamos National Security, LLC. This material was    !
! produced under U.S. Government contract DE-AC52-06NA25396 for Los Alamos !
! National Laboratory (LANL), which is operated by Los Alamos National     !
! Security, LLC for the U.S. Department of Energy. The U.S. Government has !
! rights to use, reproduce, and distribute this software.  NEITHER THE     !
! GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY,     !
! EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS         !
! SOFTWARE.  If software is modified to produce derivative works, such     !
! modified software should be clearly marked, so as not to confuse it      !
! with the version available from LANL.                                    !
!                                                                          !
! Additionally, this program is free software; you can redistribute it     !
! and/or modify it under the terms of the GNU General Public License as    !
! published by the Free Software Foundation; version 2.0 of the License.   !
! Accordingly, this program is distributed in the hope that it will be     !
! useful, but WITHOUT ANY WARRANTY; without even the implied warranty of   !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General !
! Public License for more details.                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE SPINRHOEVECS

  USE CONSTANTS_MOD
  USE SETUPARRAY
  USE DIAGARRAY
  USE SPINARRAY
  USE MYPRECISION

  IMPLICIT NONE

  INTEGER :: I, J, K
  INTEGER :: ITER, BREAKLOOP
  REAL(LATTEPREC) :: OCCERROR , OCCUP, OCCDOWN
  REAL(LATTEPREC) :: PREVERROR, PREVERROR2, PREVERROR3
  REAL(LATTEPREC) :: FDIRACUP, FDIRACDOWN
  REAL(LATTEPREC) :: SHIFTCP, MAXSHIFT = ONE
  REAL(LATTEPREC) :: UPFDARG, DOWNFDARG, DFDIRACUP, DFDIRACDOWN

  RHOUP = ZERO
  RHODOWN = ZERO
  
  ITER = 0
  BREAKLOOP = 0

  PREVERROR = ZERO
  PREVERROR2 = ZERO
  PREVERROR3 = ZERO

  !
  ! For a finite electronic temperature
  !
  
  IF (KBT .GT. 0.000001) THEN

     DO WHILE ( BREAKLOOP .EQ. 0 ) 
        
        ITER = ITER + 1
        
        IF (ITER .GT. 100) THEN
           WRITE(6,*) "NR scheme to find the chemical potential has broken"
           STOP
        ENDIF
  
        OCCUP = ZERO
        OCCDOWN = ZERO
        DFDIRACUP = ZERO
        DFDIRACDOWN = ZERO
        
        DO I = 1, HDIM
        
           UPFDARG = (UPEVALS(I) - CHEMPOT)/KBT
           DOWNFDARG = (DOWNEVALS(I) - CHEMPOT)/KBT
           
           IF (UPFDARG .GT. 46.0D0) THEN
              FDIRACUP = ZERO
           ELSEIF (UPFDARG .LT. -46.0D0) THEN
              FDIRACUP = ONE
           ELSE 
              FDIRACUP = ONE/(ONE + EXP(UPFDARG))
           ENDIF

           OCCUP = OCCUP + FDIRACUP

           IF (DOWNFDARG .GT. 46.0D0) THEN
              FDIRACDOWN = ZERO
           ELSEIF (DOWNFDARG .LT. -46.0D0) THEN
              FDIRACDOWN = ONE
           ELSE
              FDIRACDOWN = ONE/(ONE + EXP(DOWNFDARG))
           ENDIF
           
           OCCDOWN = OCCDOWN + FDIRACDOWN
           
           DFDIRACUP = DFDIRACUP + EXP(UPFDARG)*FDIRACUP*FDIRACUP/KBT
           DFDIRACDOWN = DFDIRACDOWN + EXP(DOWNFDARG)*FDIRACDOWN*FDIRACDOWN/KBT
           
        ENDDO
        
        SHIFTCP = (TOTNE - OCCUP - OCCDOWN)/(DFDIRACUP + DFDIRACDOWN)
        
        IF (ABS(SHIFTCP) .GT. MAXSHIFT) THEN
           SHIFTCP = SIGN(MAXSHIFT, SHIFTCP)
        ENDIF
        
        CHEMPOT = CHEMPOT + SHIFTCP
        
        PREVERROR3 = PREVERROR2
        PREVERROR2 = PREVERROR
        PREVERROR = OCCERROR
        OCCERROR = ABS(TOTNE - OCCUP - OCCDOWN)
        
!        PRINT*, PREVERROR, PREVERROR2, PREVERROR3
!        PRINT*, ITER, OCCERROR, CHEMPOT
        
        IF (LATTEPREC .EQ. KIND(0.0D0)) THEN
           
           ! Convergence criterion for double precision...                        
           
           IF (OCCERROR .LT. BREAKTOL) THEN
              BREAKLOOP = 1
           ENDIF
           
        ELSE
           
           ! ...and for single precison                                           
           
           IF (OCCERROR .EQ. PREVERROR .OR. &
                OCCERROR .EQ. PREVERROR2 .OR. &
                OCCERROR .EQ. PREVERROR3 .OR. ITER .EQ. 25) THEN
              
              BREAKLOOP = 1
              
           ENDIF
           
        ENDIF
     ENDDO
     
     ! Now we have the chemical potential we can build the density matrices
     
     DO I = 1, HDIM
        
        UPFDARG = (UPEVALS(I) - CHEMPOT)/KBT
        
        IF (UPFDARG .GT. 46.0D0) THEN
           FDIRACUP = ZERO
        ELSEIF (UPFDARG .LT. -46.0D0) THEN
           FDIRACUP = ONE
        ELSE
           FDIRACUP = ONE/(ONE + EXP(UPFDARG))
        ENDIF
        
        DOWNFDARG = (DOWNEVALS(I) - CHEMPOT)/KBT
        
        IF (DOWNFDARG .GT. 46.0D0) THEN
           FDIRACDOWN = ZERO
        ELSEIF (DOWNFDARG .LT. -46.0D0) THEN
           FDIRACDOWN = ONE
        ELSE
           FDIRACDOWN = ONE/(ONE + EXP(DOWNFDARG))
        ENDIF
        
        DO J = 1, HDIM
           
           SCALEDEVECSUP(J,I) = UPEVECS(J,I)*FDIRACUP
           SCALEDEVECSDOWN(J,I) = DOWNEVECS(J,I)*FDIRACDOWN
           
        ENDDO
     ENDDO

!$OMP PARALLEL DO DEFAULT(NONE) SCHEDULE(GUIDED) &
!$OMP SHARED(HDIM, RHOUP, RHODOWN, SCALEDEVECSUP, SCALEDEVECSDOWN, UPEVECS, DOWNEVECS) &
!$OMP PRIVATE(I, J, K)
     
     DO I = 1, HDIM
        DO J = 1, HDIM
           DO K = 1, HDIM
              
              RHOUP(J,I) = RHOUP(J,I) + SCALEDEVECSUP(J,K)*UPEVECS(I,K)
              
              RHODOWN(J,I) = RHODOWN(J,I) + &
                   SCALEDEVECSDOWN(J,K)*DOWNEVECS(I,K)
              
           ENDDO
        ENDDO
     ENDDO

!$OMP END PARALLEL DO     

  ELSE  ! For zero electronic temperature

     PRINT*, "spin-polarized at T_e = 0 still to come"
     STOP

  ENDIF
     
  RETURN

END SUBROUTINE SPINRHOEVECS
